package tools.descartes.prisma.agent.wildfly.extension;

import static org.jboss.as.controller.descriptions.ModelDescriptionConstants.SUBSYSTEM;

import java.util.List;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;

import org.jboss.as.controller.Extension;
import org.jboss.as.controller.ExtensionContext;
import org.jboss.as.controller.PathAddress;
import org.jboss.as.controller.PathElement;
import org.jboss.as.controller.PersistentResourceXMLDescription;
import org.jboss.as.controller.SubsystemRegistration;
import org.jboss.as.controller.descriptions.StandardResourceDescriptionResolver;
import org.jboss.as.controller.operations.common.GenericSubsystemDescribeHandler;
import org.jboss.as.controller.parsing.ExtensionParsingContext;
import org.jboss.as.controller.persistence.SubsystemMarshallingContext;
import org.jboss.as.controller.registry.ManagementResourceRegistration;
import org.jboss.dmr.ModelNode;
import org.jboss.staxmapper.XMLElementReader;
import org.jboss.staxmapper.XMLElementWriter;
import org.jboss.staxmapper.XMLExtendedStreamReader;
import org.jboss.staxmapper.XMLExtendedStreamWriter;

import tools.descartes.prisma.agent.wildfly.filters.FilterDefinition;
import tools.descartes.prisma.agent.wildfly.parameters.ParameterCharacterizationDefinition;


/**
 * @author <a href="kabir.khan@jboss.com">Kabir Khan</a>
 */
public class SubsystemExtension implements Extension {

    /**
     * The name space used for the {@code substystem} element
     */
    public static final String NAMESPACE = "urn:prisma.agent.wildfly:1.0";

    /**
     * The name of our subsystem within the model.
     */
    public static final String SUBSYSTEM_NAME = "prisma";

    /**
     * The parser used for parsing our subsystem
     */
    private final SubsystemParser parser = new SubsystemParser();

    protected static final PathElement SUBSYSTEM_PATH = PathElement.pathElement(SUBSYSTEM, SUBSYSTEM_NAME);
    private static final String RESOURCE_NAME = SubsystemExtension.class.getPackage().getName() + ".LocalDescriptions";
    
    public static final String FILTER = "filter";
    public static final String PARAM_CHARACTERIZATION = "parameter-characterization";
    
    public static final String COMPONENT_PATTERN = "component-pattern";
    public static final String INTERFACE_PATTERN = "interface-pattern";
    public static final String OPERATION_PATTERN = "operation-pattern";    
    public static final String PARAMETER = "parameter";
    public static final String CHARACTERIZATION_TYPE = "type";


    public static StandardResourceDescriptionResolver getResourceDescriptionResolver(final String keyPrefix) {
        String prefix = SUBSYSTEM_NAME + (keyPrefix == null ? "" : "." + keyPrefix);
        return new StandardResourceDescriptionResolver(prefix, RESOURCE_NAME, SubsystemExtension.class.getClassLoader(), true, false);
    }

    @Override
    public void initializeParsers(ExtensionParsingContext context) {
        context.setSubsystemXmlMapping(SUBSYSTEM_NAME, NAMESPACE, parser);
    }


    @Override
    public void initialize(ExtensionContext context) {
        final SubsystemRegistration subsystem = context.registerSubsystem(SUBSYSTEM_NAME, 1, 0);
        final ManagementResourceRegistration registration = subsystem.registerSubsystemModel(SubsystemDefinition.INSTANCE);
        registration.registerOperationHandler(GenericSubsystemDescribeHandler.DEFINITION, GenericSubsystemDescribeHandler.INSTANCE);

        subsystem.registerXMLElementWriter(parser);
    }

    /**
     * The subsystem parser, which uses stax to read and write to and from xml
     */
    private static class SubsystemParser implements XMLStreamConstants, XMLElementReader<List<ModelNode>>, XMLElementWriter<SubsystemMarshallingContext> {

        private static PersistentResourceXMLDescription xmlDescription;
        
        static {
        	xmlDescription = PersistentResourceXMLDescription.builder(SubsystemDefinition.INSTANCE)
        			.addChild(PersistentResourceXMLDescription.builder(FilterDefinition.INSTANCE)
    							.addAttribute(FilterDefinition.COMPONENT_PATTERN)
    							.addAttribute(FilterDefinition.INTERFACE_PATTERN)
    							.addAttribute(FilterDefinition.OPERATION_PATTERN)
        					)
        			.addChild(PersistentResourceXMLDescription.builder(ParameterCharacterizationDefinition.INSTANCE)
        						.addAttribute(ParameterCharacterizationDefinition.INTERFACE)
        						.addAttribute(ParameterCharacterizationDefinition.OPERATION)
        						.addAttribute(ParameterCharacterizationDefinition.PARAMETER)
        						.addAttribute(ParameterCharacterizationDefinition.TYPE)
        					).build();
        }
    	
    	/**
         * {@inheritDoc}
         */
        @Override
        public void writeContent(XMLExtendedStreamWriter writer, SubsystemMarshallingContext context) throws XMLStreamException {
        	ModelNode node = new ModelNode();        	
        	node.get(SubsystemDefinition.INSTANCE.getPathElement().getKeyValuePair()).set(context.getModelNode());
        	xmlDescription.persist(writer, node, SubsystemExtension.NAMESPACE);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public void readElement(XMLExtendedStreamReader reader, List<ModelNode> list) throws XMLStreamException {
        	xmlDescription.parse(reader, PathAddress.EMPTY_ADDRESS, list);
        }
    }

}
