package tools.descartes.prisma.agent.wildfly.service;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.UUID;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.jboss.logging.Logger;
import org.jboss.msc.service.Service;
import org.jboss.msc.service.ServiceName;
import org.jboss.msc.service.StartContext;
import org.jboss.msc.service.StartException;
import org.jboss.msc.service.StopContext;
import org.jboss.msc.value.InjectedValue;

import edu.kit.ipd.descartes.mm.applicationlevel.parameterdependencies.ModelVariableCharacterizationType;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyConnector;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyContext;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.BasicComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.ComposedStructure;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.CompositeComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.Interface;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceProvidingRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceRequiringRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.ProvidingDelegationConnector;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RequiringDelegationConnector;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.Signature;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.AbstractAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ComponentInternalBehavior;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ExternalCall;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ExternalCallAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.FineGrainedBehavior;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ForkAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.InternalAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.LoopAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.LoopIterationCount;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ResourceDemand;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ServicebehaviorFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.system.SystemFactory;
import edu.kit.ipd.descartes.mm.deployment.DeploymentContext;
import edu.kit.ipd.descartes.mm.deployment.DeploymentFactory;
import edu.kit.ipd.descartes.mm.resourcelandscape.RuntimeEnvironment;
import edu.kit.ipd.descartes.mm.runtimeenvironmentclasses.RuntimeEnvironmentClasses;
import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.agent.wildfly.filters.OperationPredicate;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.agent.AgentController;
import tools.descartes.prisma.core.agent.RemoveFilter;
import tools.descartes.prisma.core.impl.CDOModelRepository;
import tools.descartes.prisma.core.util.ModelSkeletonUtil;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;
import tools.descartes.prisma.model.skeleton.SkeletonFactory;

public class ModelExtractionService extends AgentController implements Service<ModelExtractionService> {

	private class CallPaths {
		private Map<String, List<String>> paths = new HashMap<>();

		public void addPath(String interf, String assemblyCtx) {
			List<String> a = paths.get(interf);
			if (a == null) {
				a = new LinkedList<>();
				paths.put(interf, a);
			}
			a.add(assemblyCtx);
		}

		public int getIndex(String interf, String assemblyCtx) {
			List<String> a = paths.get(interf);
			if (a != null) {
				int idx = 0;
				for (String cur : a) {
					if (cur.equals(assemblyCtx)) {
						return idx;
					}
					idx++;
				}
			}
			return -1;
		}
	}

	private static final Logger log = Logger.getLogger(ModelExtractionService.class);

	private InjectedValue<MonitoringService> monitoring = new InjectedValue<>();

	private List<InjectedValue<FilterService>> filters = new CopyOnWriteArrayList<>();

	public static final ServiceName SERVICE_NAME = ServiceName.JBOSS.append("modelextraction", "modelextraction");

	private static InetAddress localHost;
	static {
		try {
			localHost = InetAddress.getLocalHost();
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}
	}

	public static final ModelRepository repository = new CDOModelRepository("10.0.1.148");

	public static final ModelExtractionService INSTANCE = new ModelExtractionService();

	private final List<OperationPredicate> operationFilters = new LinkedList<>();

	private ModelSkeleton skeleton;

	private RuntimeEnvironment wildFlyContainer;

	private Map<String, CallPaths> callpaths = new HashMap<>();

	private Queue<AssemblyContext> undeployedAssemblyContexts = new LinkedList<>();

	private Queue<Object> updates = new ConcurrentLinkedQueue<>();

	public ModelExtractionService() {
		super(repository, "application/specj", localHost.getHostName().toString() + "_" + UUID.randomUUID().toString());
	}

	@Override
	public ModelExtractionService getValue() throws IllegalStateException, IllegalArgumentException {
		return this;
	}

	@Override
	public void start(StartContext arg0) throws StartException {
		for (InjectedValue<FilterService> filter : filters) {
			operationFilters.add(filter.getValue().getPredicate());
		}

		try {
			localHost = InetAddress.getLocalHost();
		} catch (UnknownHostException e1) {
			e1.printStackTrace();
		}

		skeleton = SkeletonFactory.eINSTANCE.createModelSkeleton();
		skeleton.setRepository(RepositoryFactory.eINSTANCE.createRepository());
		skeleton.setSystem(SystemFactory.eINSTANCE.createSystem());
		skeleton.setDeployment(DeploymentFactory.eINSTANCE.createDeployment());

		setRemoveFilter(new RemoveFilter() {
			@Override
			public boolean isRemovedOnShutdown(EObject object) {
				if (object instanceof DeploymentContext) {
					// Deployment contexts are always obsolete when wildfly is
					// shut down
					return true;
				}
				if (object instanceof AssemblyContext || object instanceof AssemblyConnector
						|| object instanceof ProvidingDelegationConnector
						|| object instanceof RequiringDelegationConnector || object instanceof InterfaceProvidingRole
						|| object instanceof InterfaceRequiringRole) {
					// These elements are only removed if they are contained in
					// a System or Subsystem. CompositeComponents should stay as
					// they are.
					return !(object.eContainer() instanceof CompositeComponent)
							&& !(object.eContainer() instanceof BasicComponent);
				}
				return false;
			}
		});

		try {
			start().get();
			schedule(new ModelUpdater(), 1, TimeUnit.MINUTES);
		} catch (InterruptedException | ExecutionException e) {
			log.error("Error stopping agent.", e);
		}
	}

	@Override
	public void stop(StopContext arg0) {
		try {
			stop().get();
		} catch (InterruptedException | ExecutionException e) {
			log.error("Error stopping agent.", e);
		}
	}

	public List<InjectedValue<FilterService>> getFilters() {
		return filters;
	}

	public void newComponentRecord(ComponentRecord record) {
		updates.add(record);
	}

	public void newOperationRecord(OperationRecord record) {
		updates.add(record);
	}

	public void newInvocationRecord(InvocationRecord record) {
		updates.add(record);
	}

	@Override
	protected void notifyAddDelegatedElement(EObject newElement, Map<String, String> additionalProperties) {
		if (newElement instanceof RuntimeEnvironment) {
			RuntimeEnvironment env = (RuntimeEnvironment) newElement;
			if (env.getOfClass() == RuntimeEnvironmentClasses.OS_VM
					&& localHost.getHostAddress().equals(additionalProperties.get("container.ip"))) {
				wildFlyContainer = env;
			}
		}
	}

	private class ModelUpdater implements Runnable {

		@Override
		public void run() {
			try {
				Object current;

				boolean changed = false;

				while ((current = updates.poll()) != null) {
					if (current instanceof InvocationRecord) {
						handleInvocation((InvocationRecord) current);
						changed = true;
					}
				}

				if (changed) {
					apply(skeleton);
				}
			} catch (Throwable ex) {
				log.error("Error updating skeleton.", ex);
			}
		}

		private void handleInvocation(InvocationRecord record) {
			InvocationRecord caller = record.getOperation().getCaller();
			if (!isFiltered(record)) {
				if (caller != null) {
					if (!isFiltered(caller)) {
						// We have information on the incoming calls as well as
						// the outgoing calls.
						addComponent(caller, record);
					}
				} else {
					// no information about outgoing calls.
					addComponent(record, null);
				}
			}
		}
	}

	private boolean isFiltered(InvocationRecord record) {
		boolean filtered = false;
		String componentName = record.getOperation().getComponent().getComponentName();
		for (OperationPredicate filter : operationFilters) {
			if (filter.apply(componentName, record.getOperation().getInterfaceName(),
					record.getOperation().getOperationName())) {
				filtered = true;
				break;
			}
		}
		return filtered;
	}

	private void addComponent(InvocationRecord incomingIrec, InvocationRecord outgoingIrec) {
		ComponentRecord incomingCrec = incomingIrec.getOperation().getComponent();
		String incomingInterfaceName = incomingIrec.getOperation().getInterfaceName();
		boolean isRemoteIncomingCall = !incomingIrec.getOperation().getProtocol().equals("java");
		ComponentRecord outgoingCrec = null;
		String outgoingInterfaceName = null;
		if (outgoingIrec != null) {
			outgoingCrec = outgoingIrec.getOperation().getComponent();
			outgoingInterfaceName = outgoingIrec.getOperation().getInterfaceName();
		}

		// 1. add the incoming interface to the repository
		Interface incomingInterface = RepositoryFactory.eINSTANCE.createInterface();
		incomingInterface.setName(incomingInterfaceName);
		incomingInterface = ModelSkeletonUtil.insert(skeleton.getRepository().getInterfaces(), incomingInterface);

		Signature incomingSignature = RepositoryFactory.eINSTANCE.createSignature();
		incomingSignature.setName(getSignatureName(incomingIrec));
		incomingSignature = ModelSkeletonUtil.insert(incomingInterface.getSignatures(), incomingSignature);

		// 2. add the outgoing interface to the repository (if applicable)
		Interface outgoingInterface = null;
		Signature outgoingSignature = null;
		if (outgoingIrec != null) {
			outgoingInterface = RepositoryFactory.eINSTANCE.createInterface();
			outgoingInterface.setName(outgoingInterfaceName);
			outgoingInterface = ModelSkeletonUtil.insert(skeleton.getRepository().getInterfaces(), outgoingInterface);
			outgoingSignature = RepositoryFactory.eINSTANCE.createSignature();
			outgoingSignature.setName(getSignatureName(outgoingIrec));
			outgoingSignature = ModelSkeletonUtil.insert(outgoingInterface.getSignatures(), outgoingSignature);
		}

		// 3. Create the implementation component for the caller component
		BasicComponent implementationComponent = RepositoryFactory.eINSTANCE.createBasicComponent();
		implementationComponent.setName(incomingCrec.getComponentClass());
		implementationComponent = (BasicComponent) ModelSkeletonUtil.insert(skeleton.getRepository().getComponents(),
				implementationComponent);
		String providingPortName = getInterfaceProvidingRoleName(incomingIrec, "");
		InterfaceProvidingRole callerProvidingRole = RepositoryFactory.eINSTANCE.createInterfaceProvidingRole();
		callerProvidingRole.setName(providingPortName);
		callerProvidingRole.setInterface(incomingInterface);
		callerProvidingRole = ModelSkeletonUtil.insert(implementationComponent.getInterfaceProvidingRoles(),
				callerProvidingRole);
		InterfaceRequiringRole callerRequiringRole = null;
		int callerRoleIdx = 0;
		if (outgoingInterface != null) {
			// Here we check whether there are external calls for the same
			// interface
			// but to different target assembly contexts. If yes, we need to
			// distinguish the two requiring ports
			callerRoleIdx = getRequiringPortIndex(incomingCrec, outgoingIrec.getOperation());
			String roleName = getInterfaceRequiringRoleName(outgoingIrec, "");
			if (callerRoleIdx > 0) {
				roleName = roleName + "/" + callerRoleIdx;
			}
			callerRequiringRole = RepositoryFactory.eINSTANCE.createInterfaceRequiringRole();
			callerRequiringRole.setName(roleName);
			callerRequiringRole.setInterface(outgoingInterface);
			callerRequiringRole = ModelSkeletonUtil.insert(implementationComponent.getInterfaceRequiringRoles(),
					callerRequiringRole);
		}
		FineGrainedBehavior behavior = createComponentInternalBehavior(implementationComponent, callerProvidingRole,
				incomingSignature);
		if (callerRequiringRole != null) {
			addExternalCall(incomingCrec.getComponentClass(), implementationComponent, behavior, callerRequiringRole,
					outgoingSignature);
		}
		RepositoryComponent callerComponent = implementationComponent;

		// 4. Now add the called component (if applicable)
		InterfaceProvidingRole calledProvidingRole = null;
		RepositoryComponent calledComponent = null;
		boolean differentApplication = false;
		if (callerRequiringRole != null) {
			differentApplication = isDifferentApplication(incomingCrec, outgoingCrec);
			calledComponent = createCalledComponent(outgoingCrec, differentApplication);
			String calledProvingRoleName;
			if (calledComponent instanceof CompositeComponent) {
				calledProvingRoleName = getInterfaceProvidingRoleName(outgoingIrec,
						outgoingIrec.getOperation().getComponent().getComponentName());
			} else {
				calledProvingRoleName = getInterfaceProvidingRoleName(outgoingIrec, "");
			}

			calledProvidingRole = RepositoryFactory.eINSTANCE.createInterfaceProvidingRole();
			calledProvidingRole.setName(calledProvingRoleName);
			calledProvidingRole.setInterface(outgoingInterface);
			calledProvidingRole = ModelSkeletonUtil.insert(calledComponent.getInterfaceProvidingRoles(),
					calledProvidingRole);
		}

		// 5. Now wrap the implementation in a composite component (if
		// necessary)
		CompositeComponent compositeComponent = createCompositeComponent(incomingCrec, callerComponent);
		if (compositeComponent != null) {

			// Create an assembly context for embedding the component and
			// connect its interface providing role with a role on the composite
			// component
			AssemblyContext assembly = addAssemblyContext(compositeComponent, incomingCrec, implementationComponent);

			if (isRemoteIncomingCall) {
				String providingRemotePortName = getInterfaceProvidingRoleName(incomingIrec,
						incomingCrec.getComponentName());
				InterfaceProvidingRole remoteProvidingRole = RepositoryFactory.eINSTANCE.createInterfaceProvidingRole();
				remoteProvidingRole.setName(providingRemotePortName);
				remoteProvidingRole.setInterface(incomingInterface);
				remoteProvidingRole = ModelSkeletonUtil.insert(compositeComponent.getInterfaceProvidingRoles(),
						remoteProvidingRole);

				ProvidingDelegationConnector providingConnector = RepositoryFactory.eINSTANCE
						.createProvidingDelegationConnector();
				providingConnector.setName(assembly.getName());
				providingConnector.setOuterInterfaceProvidingRole(remoteProvidingRole);
				providingConnector.setInnerInterfaceProvidingRole(callerProvidingRole);
				providingConnector.setAssemblyContext(assembly);
				ModelSkeletonUtil.insert(compositeComponent.getProvidingDelegationConnectors(), providingConnector);

				// IMPORTANT: from now on use this as the target of assembly
				// connectors.
				callerProvidingRole = remoteProvidingRole;
			}

			InterfaceRequiringRole remoteRequiringRole = null;
			if (callerRequiringRole != null) {
				if (differentApplication) {
					// Only add a interface requiring role to the composite
					// component if it is a remote call, i.e., a call to another
					// composite component.
					String requiringRemotePortName = getInterfaceRequiringRoleName(outgoingIrec,
							incomingCrec.getComponentName());
					if (callerRoleIdx > 0) {
						requiringRemotePortName = requiringRemotePortName + "/" + callerRoleIdx;
					}
					remoteRequiringRole = RepositoryFactory.eINSTANCE.createInterfaceRequiringRole();
					remoteRequiringRole.setName(requiringRemotePortName);
					remoteRequiringRole.setInterface(outgoingInterface);
					remoteRequiringRole = ModelSkeletonUtil.insert(compositeComponent.getInterfaceRequiringRoles(),
							remoteRequiringRole);

					RequiringDelegationConnector requiringConnector = RepositoryFactory.eINSTANCE
							.createRequiringDelegationConnector();
					requiringConnector.setName(assembly.getName());
					requiringConnector.setOuterInterfaceRequiringRole(remoteRequiringRole);
					requiringConnector.setInnerInterfaceRequiringRole(callerRequiringRole);
					requiringConnector.setAssemblyContext(assembly);
					ModelSkeletonUtil.insert(compositeComponent.getRequiringDelegationConnectors(), requiringConnector);
				} else {
					// Local call: Also add an assembly context for the called
					// component within the composite component
					// and connect it with an assembly connector.
					AssemblyContext calledAssembly = addAssemblyContext(compositeComponent, outgoingCrec,
							calledComponent);

					AssemblyConnector connector = RepositoryFactory.eINSTANCE.createAssemblyConnector();
					connector.setName(assembly.getName() + " -> " + calledAssembly.getName());
					connector.setInterfaceProvidingRole(calledProvidingRole);
					connector.setInterfaceRequiringRole(callerRequiringRole);
					connector.setProvidingAssemblyContext(calledAssembly);
					connector.setRequiringAssemblyContext(assembly);
					ModelSkeletonUtil.insert(compositeComponent.getAssemblyConnectors(), connector);
				}
			}
			callerRequiringRole = remoteRequiringRole;
			callerComponent = compositeComponent;
		}

		// 6. Now create the assembly contexts
		AssemblyContext callerContext = addAssemblyContext(skeleton.getSystem(), incomingCrec, callerComponent);
		if (callerRequiringRole != null) {
			AssemblyContext calledContext = addAssemblyContext(skeleton.getSystem(), outgoingCrec, calledComponent);

			AssemblyConnector connector = RepositoryFactory.eINSTANCE.createAssemblyConnector();
			connector.setName(callerContext.getName() + " -> " + calledContext.getName());
			connector.setInterfaceProvidingRole(calledProvidingRole);
			connector.setInterfaceRequiringRole(callerRequiringRole);
			connector.setProvidingAssemblyContext(calledContext);
			connector.setRequiringAssemblyContext(callerContext);
			ModelSkeletonUtil.insert(skeleton.getSystem().getAssemblyConnectors(), connector);
		}

		// 7. Expose the provided remote interfaces as system interfaces (if
		// this port is marked as system interface)
		String systemProvidingRoleName = getInterfaceProvidingRoleName(incomingIrec, callerContext.getName());
		// Do not create a new providing/requiring role for the system, only use
		// the existing providing/requiring role.
		// This way a user can control which interfaces are exposed at the
		// system level by creating them manually.
		InterfaceProvidingRole systemProvidingRole = RepositoryFactory.eINSTANCE.createInterfaceProvidingRole();
		systemProvidingRole.setName(systemProvidingRoleName);
		systemProvidingRole.setInterface(incomingInterface);
		systemProvidingRole = ModelSkeletonUtil.find(
				getContainerScope().getSystemScope().getSystem().getInterfaceProvidingRoles(), systemProvidingRole);
		if (systemProvidingRole != null) {
			skeleton.getSystem().getInterfaceProvidingRoles()
					.add(ModelSkeletonUtil.adopt(this, systemProvidingRole));

			ProvidingDelegationConnector systemProvidingConnector = RepositoryFactory.eINSTANCE
					.createProvidingDelegationConnector();
			systemProvidingConnector.setName(callerContext.getName());
			systemProvidingConnector.setOuterInterfaceProvidingRole(systemProvidingRole);
			systemProvidingConnector.setInnerInterfaceProvidingRole(callerProvidingRole);
			systemProvidingConnector.setAssemblyContext(callerContext);
			ModelSkeletonUtil.insert(skeleton.getSystem().getProvidingDelegationConnectors(), systemProvidingConnector);
		}

		if (callerRequiringRole != null) {
			String systemRequiringRoleName = getInterfaceRequiringRoleName(outgoingIrec, callerContext.getName());
			InterfaceRequiringRole systemRequiringRole = RepositoryFactory.eINSTANCE.createInterfaceRequiringRole();
			systemRequiringRole.setName(systemRequiringRoleName);
			systemRequiringRole.setInterface(outgoingInterface);
			systemRequiringRole = ModelSkeletonUtil.find(
					getContainerScope().getSystemScope().getSystem().getInterfaceRequiringRoles(), systemRequiringRole);
			if (systemRequiringRole != null) {
				skeleton.getSystem().getInterfaceRequiringRoles()
						.add(ModelSkeletonUtil.adopt(this, systemRequiringRole));

				RequiringDelegationConnector systemRequiringConnector = RepositoryFactory.eINSTANCE
						.createRequiringDelegationConnector();
				systemRequiringConnector.setName(callerContext.getName());
				systemRequiringConnector.setInnerInterfaceRequiringRole(callerRequiringRole);
				systemRequiringConnector.setOuterInterfaceRequiringRole(systemRequiringRole);
				systemRequiringConnector.setAssemblyContext(callerContext);
				ModelSkeletonUtil.insert(skeleton.getSystem().getRequiringDelegationConnectors(),
						systemRequiringConnector);
			}
		}

		// 8. Remember to deploy this context (only for the caller component)
		undeployedAssemblyContexts.add(callerContext);
	}

	private FineGrainedBehavior createComponentInternalBehavior(BasicComponent implementationComponent,
			InterfaceProvidingRole implementationProvidingRole, Signature signature) {
		FineGrainedBehavior behavior = ServicebehaviorFactory.eINSTANCE.createFineGrainedBehavior();
		behavior.setDescribedSignature(signature);
		behavior.setInterfaceProvidingRole(implementationProvidingRole);
		behavior = ModelSkeletonUtil.insert(implementationComponent.getFineGrainedBehavior(), behavior);
		if (behavior.getBehavior() == null) {
			behavior.setBehavior(ServicebehaviorFactory.eINSTANCE.createComponentInternalBehavior());
		}
		if (!isJMSComponent(implementationComponent.getName())) {
			if (behavior.getBehavior().getActions().isEmpty()) {
				InternalAction internalAction = ServicebehaviorFactory.eINSTANCE.createInternalAction();
				ResourceDemand cpuDemand = ServicebehaviorFactory.eINSTANCE.createResourceDemand();
				cpuDemand.setName("CPU");
				cpuDemand.setCharacterization(ModelVariableCharacterizationType.EMPIRICAL);
				cpuDemand.setResourceType(
						ModelSkeletonUtil.adopt(this, getContainerScope().getSystemScope().getCpuResourceType()));
				internalAction.getResourceDemand().add(cpuDemand);
				ModelSkeletonUtil.insert(behavior.getBehavior().getActions(), internalAction);
			}
		}
		return behavior;
	}

	private AssemblyContext addAssemblyContext(ComposedStructure parent, ComponentRecord componentRecord,
			RepositoryComponent component) {
		AssemblyContext context = RepositoryFactory.eINSTANCE.createAssemblyContext();
		StringBuilder contextName = new StringBuilder();
		if (!(parent instanceof CompositeComponent)) {
			// In order to uniquely identify the system level assembly contexts,
			// we also include the server node name (e.g., in case of
			// replicated tiers)
			contextName.append(componentRecord.getServerNode());
			contextName.append("/");
		}
		if (component instanceof CompositeComponent) {
			// This component is the container for more basic components
			// use application and module in name
			contextName.append(componentRecord.getApplication());
			contextName.append("/");
			contextName.append(componentRecord.getModule());
			contextName.append("/");
		} else {
			contextName.append(componentRecord.getComponentName());
		}
		context.setName(contextName.toString());
		context.setEncapsulatedComponent(component);
		return ModelSkeletonUtil.insert(parent.getAssemblyContexts(), context);
	}

	private void addExternalCall(String componentName, BasicComponent implementationComponent,
			FineGrainedBehavior behavior, InterfaceRequiringRole requiringRole, Signature outgoingSignature) {
		ExternalCallAction callAction = ServicebehaviorFactory.eINSTANCE.createExternalCallAction();
		ExternalCall call = ServicebehaviorFactory.eINSTANCE.createExternalCall();
		call.setName(requiringRole.getName() + "#" + outgoingSignature.getName());
		call.setSignature(outgoingSignature);
		call.setInterfaceRequiringRole(requiringRole);
		callAction.setExternalCall(call);

		if (isJMSComponent(componentName)) {
			if (behavior.getBehavior().getActions().isEmpty()) {
				ForkAction fork = ServicebehaviorFactory.eINSTANCE.createForkAction();
				fork.setWithSynchronizationBarrier(false);
				ComponentInternalBehavior forkedBehavior = ServicebehaviorFactory.eINSTANCE
						.createComponentInternalBehavior();
				forkedBehavior.getActions().add(callAction);
				fork.getForkedBehaviors().add(forkedBehavior);
				behavior.getBehavior().getActions().add(fork);
			}
		} else {
			LoopAction loop = ServicebehaviorFactory.eINSTANCE.createLoopAction();
			LoopIterationCount iterationCount = ServicebehaviorFactory.eINSTANCE.createLoopIterationCount();
			iterationCount.setName(call.getName());
			iterationCount.setCharacterization(ModelVariableCharacterizationType.EMPIRICAL);
			loop.setLoopIterationCount(iterationCount);
			loop.setLoopBodyBehavior(ServicebehaviorFactory.eINSTANCE.createComponentInternalBehavior());
			loop.getLoopBodyBehavior().getActions().add(callAction);

			// We assume that the external calls are inserted in
			// alphabetical order.
			EList<AbstractAction> actions = behavior.getBehavior().getActions();
			for (int i = 0; i < actions.size(); i++) {
				if (actions.get(i) instanceof LoopAction) {
					LoopAction curLoop = (LoopAction) actions.get(i);
					int compareResult = curLoop.getLoopIterationCount().getName().compareTo(call.getName());
					if (compareResult == 0) {
						// Ok, there is already such an external call in the
						// behavior.
						return;
					} else if (compareResult > 0) {
						actions.add(i, loop);
						return;
					}
				}
			}
			// The previous loop did not return, i.e., the new action is added
			// at the back.
			actions.add(loop);
		}
	}

	private CompositeComponent createCompositeComponent(ComponentRecord crec,
			RepositoryComponent implementationComponent) {
		String compositeName = getCompositeComponentName(crec);
		if (compositeName.length() > 0) {
			CompositeComponent compositeComponent = RepositoryFactory.eINSTANCE.createCompositeComponent();
			compositeComponent.setName(compositeName.toString());
			return (CompositeComponent) ModelSkeletonUtil.insert(skeleton.getRepository().getComponents(),
					compositeComponent);
		}
		return null;
	}

	private RepositoryComponent createCalledComponent(ComponentRecord callee,
			boolean remoteCall) {
		// check whether this call goes to a component in the same composite or
		// not
		if (remoteCall) {
			String calleeCompositeName = getCompositeComponentName(callee);
			if (calleeCompositeName.length() > 0) {
				CompositeComponent compositeComponent = RepositoryFactory.eINSTANCE.createCompositeComponent();
				compositeComponent.setName(calleeCompositeName);
				return (CompositeComponent) ModelSkeletonUtil.insert(skeleton.getRepository().getComponents(),
						compositeComponent);
			}
		}
		if (callee.getComponentClass() != null) {
			BasicComponent basicComponent = RepositoryFactory.eINSTANCE.createBasicComponent();
			basicComponent.setName(callee.getComponentClass());
			return (BasicComponent) ModelSkeletonUtil.insert(skeleton.getRepository().getComponents(), basicComponent);
		}
		throw new IllegalStateException("Could not create called component for " + callee.getComponentName());
	}

	private boolean isDifferentApplication(ComponentRecord caller, ComponentRecord callee) {
		boolean sameApplication = caller.getApplication().equals(callee.getApplication())
				&& caller.getModule().equals(callee.getModule());
		return !sameApplication;
	}

	private boolean isJMSComponent(String componentName) {
		return "javax.jms.Queue".equals(componentName) || "javax.jms.Topic".equals(componentName);
	}

	private String getCompositeComponentName(ComponentRecord crec) {
		StringBuilder compositeName = new StringBuilder();
		if (crec.getApplication() != null && !crec.getApplication().isEmpty()) {
			compositeName.append(crec.getApplication());
		}
		if (crec.getModule() != null && !crec.getModule().isEmpty()) {
			if (compositeName.length() > 0) {
				compositeName.append("/");
			}
			compositeName.append(crec.getModule());
		}
		return compositeName.toString();
	}

	private int getRequiringPortIndex(ComponentRecord incomingCrec, OperationRecord outgoingIrec) {
		CallPaths paths = callpaths.get(incomingCrec.getComponentName());
		if (paths == null) {
			paths = new CallPaths();
			callpaths.put(incomingCrec.getComponentName(), paths);
		}
		paths.addPath(outgoingIrec.getInterfaceName(), outgoingIrec.getComponent().getComponentName());
		int idx = paths.getIndex(outgoingIrec.getInterfaceName(), outgoingIrec.getComponent().getComponentName());
		if (idx < 0) {
			throw new IllegalStateException();
		}
		return idx;
	}

	public InjectedValue<MonitoringService> getMonitoringService() {
		return monitoring;
	}

	private String getSignatureName(InvocationRecord record) {
		String name = record.getOperation().getInterfaceName() + "#"
				+ record.getOperation().getOperationName();
		if (record.getOperation().areParametersMonitored()) {
			name += "_" + record.getValueBucket();
		}
		return name;
	}

	private String getInterfaceProvidingRoleName(InvocationRecord record, String assembly) {
		StringBuilder res = new StringBuilder();
		res.append(record.getOperation().getProtocol());
		res.append("://");
		if (!assembly.isEmpty()) {
			res.append(assembly);
			res.append("/");
		}
		res.append(record.getOperation().getInterfaceName());
		return res.toString();
	}

	private String getInterfaceRequiringRoleName(InvocationRecord record, String assembly) {
		StringBuilder res = new StringBuilder();
		res.append(record.getOperation().getProtocol());
		res.append("://");
		if (!assembly.isEmpty()) {
			res.append(assembly);
			res.append("/");
		}
		res.append(record.getOperation().getInterfaceName());
		return res.toString();
	}
}
