package tools.descartes.prisma.core.scopes;

import java.util.List;

import edu.kit.ipd.descartes.mm.resourcelandscape.Container;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapter;
import tools.descartes.prisma.core.agent.AgentController;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;

public class ContainerScope extends InfrastructureScope {

	private static final String CONTAINER_SUFFIX = ".containers";

	private Container parentContainer;
	private List<Container> containers;

	public ContainerScope(String name, String type, SystemGlobalScope systemScope, ModelRepository repository) {
		super(name, type, systemScope, repository);
	}

	public void setParentContainer(Container parentContainer) {
		this.parentContainer = parentContainer;
	}

	public Container getParentContainer() {
		return this.parentContainer;
	}

	@SuppressWarnings("unchecked")
	public List<Container> getContainer() {
		if (containers == null) {
			// OK, bit of a hack.
			containers = (List<Container>) (List<?>) getModelRepository().getResource(getContainerPath()).getContents();
		}
		return containers;
	}

	private String getContainerPath() {
		return getBasePath() + "/" + getName() + "/" + getName() + CONTAINER_SUFFIX;
	}

	@Override
	protected void createResources(Transaction transaction) {
		super.createResources(transaction);

		// Create the resource.
		// The containers may be added later
		getModelRepository().getOrCreateResource(transaction, getContainerPath());
	}

	@Override
	protected List<ModelSkeletonAdapter> synchronizeModelSkeletonTree(AgentController agent, Transaction transaction,
			ModelSkeleton skeleton) {
		List<ModelSkeletonAdapter> synchronizedAdapters = super.synchronizeModelSkeletonTree(agent, transaction,
				skeleton);

		// TODO: correctly implement the container matching
		for (Container curContainer : skeleton.getContainers()) {
			ModelSkeletonAdapter adapter = agent.adapt(curContainer, ModelSkeletonAdapter.class);
			adapter.startTransaction(transaction);
			adapter.synchronizeRoot(curContainer);
			synchronizedAdapters.add(adapter);
		}
		return synchronizedAdapters;
	}
}
